@@ -3,7 +3,10 @@ from __future__ import annotations
 import asyncio
 import inspect
 import logging
+import math
+import os
 import sys
+import tempfile
 import threading
 import traceback
 import types
@@ -11,10 +14,17 @@ import uuid
 import warnings
 import weakref
 from collections import defaultdict, deque
-from collections.abc import Container, Coroutine
+from collections.abc import (
+    Awaitable,
+    Callable,
+    Container,
+    Coroutine,
+    Generator,
+    Hashable,
+)
 from enum import Enum
-from functools import partial
-from typing import TYPE_CHECKING, Any, Callable, ClassVar, TypedDict, TypeVar, final
+from functools import wraps
+from typing import TYPE_CHECKING, Any, ClassVar, Literal, TypedDict, TypeVar, final
 
 import tblib
 from tlz import merge
@@ -33,20 +43,30 @@ from distributed.comm import (
     normalize_address,
     unparse_host_port,
 )
+from distributed.comm.core import Listener
 from distributed.compatibility import PeriodicCallback
-from distributed.metrics import time
+from distributed.counter import Counter
+from distributed.diskutils import WorkDir, WorkSpace
+from distributed.metrics import context_meter, time
 from distributed.system_monitor import SystemMonitor
 from distributed.utils import (
     NoOpAwaitable,
     get_traceback,
     has_keyword,
+    import_file,
+    is_python_shutting_down,
     iscoroutinefunction,
+    offload,
     recursive_to_dict,
     truncate_exception,
+    wait_for,
+    warn_on_duration,
 )
 
 if TYPE_CHECKING:
-    from typing_extensions import ParamSpec
+    from typing_extensions import ParamSpec, Self
+
+    from distributed.counter import Digest
 
     P = ParamSpec("P")
     R = TypeVar("R")
@@ -299,8 +319,54 @@ class Server:
 
     """
 
-    default_ip = ""
-    default_port = 0
+    default_ip: ClassVar[str] = ""
+    default_port: ClassVar[int] = 0
+
+    id: str
+    blocked_handlers: list[str]
+    handlers: dict[str, Callable]
+    stream_handlers: dict[str, Callable]
+    listeners: list[Listener]
+    counters: defaultdict[str, Counter]
+    deserialize: bool
+
+    local_directory: str
+
+    monitor: SystemMonitor
+    io_loop: IOLoop
+    thread_id: int
+
+    periodic_callbacks: dict[str, PeriodicCallback]
+    digests: defaultdict[Hashable, Digest] | None
+    digests_total: defaultdict[Hashable, float]
+    digests_total_since_heartbeat: defaultdict[Hashable, float]
+    digests_max: defaultdict[Hashable, float]
+
+    _last_tick: float
+    _tick_counter: int
+    _last_tick_counter: int
+    _tick_interval: float
+    _tick_interval_observed: float
+
+    _status: Status
+
+    _address: str | None
+    _listen_address: str | None
+    _host: str | None
+    _port: int | None
+
+    _comms: dict[Comm, str | None]
+
+    _ongoing_background_tasks: AsyncTaskGroup
+    _event_finished: asyncio.Event
+
+    _original_local_dir: str
+    _updated_sys_path: bool
+    _workspace: WorkSpace
+    _workdir: None | WorkDir
+
+    _startup_lock: asyncio.Lock
+    __startup_exc: Exception | None
 
     def __init__(
         self,
@@ -314,7 +380,41 @@ class Server:
         connection_args=None,
         timeout=None,
         io_loop=None,
+        local_directory=None,
+        needs_workdir=True,
     ):
+        if local_directory is None:
+            local_directory = (
+                dask.config.get("temporary-directory") or tempfile.gettempdir()
+            )
+
+        if "dask-scratch-space" not in str(local_directory):
+            local_directory = os.path.join(local_directory, "dask-scratch-space")
+
+        self._original_local_dir = local_directory
+
+        with warn_on_duration(
+            "1s",
+            "Creating scratch directories is taking a surprisingly long time. ({duration:.2f}s) "
+            "This is often due to running workers on a network file system. "
+            "Consider specifying a local-directory to point workers to write "
+            "scratch data to a local disk.",
+        ):
+            self._workspace = WorkSpace(local_directory)
+
+            if not needs_workdir:  # eg. Nanny will not need a WorkDir
+                self._workdir = None
+                self.local_directory = self._workspace.base_dir
+            else:
+                name = type(self).__name__.lower()
+                self._workdir = self._workspace.new_work_dir(prefix=f"{name}-")
+                self.local_directory = self._workdir.dir_path
+
+        self._updated_sys_path = False
+        if self.local_directory not in sys.path:
+            sys.path.insert(0, self.local_directory)
+            self._updated_sys_path = True
+
         if io_loop is not None:
             warnings.warn(
                 "The io_loop kwarg to Server is ignored and will be deprecated",
@@ -346,7 +446,6 @@ class Server:
         self._comms = {}
         self.deserialize = deserialize
         self.monitor = SystemMonitor()
-        self.counters = None
         self._ongoing_background_tasks = AsyncTaskGroup()
         self._event_finished = asyncio.Event()
 
@@ -370,24 +469,26 @@ class Server:
             else:
                 self.io_loop.profile = deque()
 
+        self.periodic_callbacks = {}
+
         # Statistics counters for various events
         try:
             from distributed.counter import Digest
 
-            self.digests = defaultdict(partial(Digest, loop=self.io_loop))
+            self.digests = defaultdict(Digest)
         except ImportError:
             self.digests = None
 
-        # In case crick is not installed, also log cumulative totals (reset at server
-        # restart) and local maximums (reset by prometheus poll)
-        self.digests_total = defaultdict(float)
-        self.digests_max = defaultdict(float)
+        # Also log cumulative totals (reset at server restart)
+        # and local maximums (reset by prometheus poll)
+        # Don't cast int metrics to float
+        self.digests_total = defaultdict(int)
+        self.digests_total_since_heartbeat = defaultdict(int)
+        self.digests_max = defaultdict(int)
 
-        from distributed.counter import Counter
-
-        self.counters = defaultdict(partial(Counter, loop=self.io_loop))
-
-        self.periodic_callbacks = {}
+        self.counters = defaultdict(Counter)
+        pc = PeriodicCallback(self._shift_counters, 5000)
+        self.periodic_callbacks["shift_counters"] = pc
 
         pc = PeriodicCallback(
             self.monitor.update,
@@ -435,6 +536,41 @@ class Server:
 
         self.__stopped = False
 
+    async def upload_file(
+        self, filename: str, data: str | bytes, load: bool = True
+    ) -> dict[str, Any]:
+        out_filename = os.path.join(self.local_directory, filename)
+
+        def func(data):
+            if isinstance(data, str):
+                data = data.encode()
+            with open(out_filename, "wb") as f:
+                f.write(data)
+                f.flush()
+                os.fsync(f.fileno())
+            return data
+
+        if len(data) < 10000:
+            data = func(data)
+        else:
+            data = await offload(func, data)
+
+        if load:
+            try:
+                import_file(out_filename)
+            except Exception as e:
+                logger.exception(e)
+                raise e
+
+        return {"status": "OK", "nbytes": len(data)}
+
+    def _shift_counters(self):
+        for counter in self.counters.values():
+            counter.shift()
+        if self.digests is not None:
+            for digest in self.digests.values():
+                digest.shift()
+
     @property
     def status(self) -> Status:
         try:
@@ -520,12 +656,12 @@ class Server:
             timeout = getattr(self, "death_timeout", None)
 
             async def _close_on_failure(exc: Exception) -> None:
-                await self.close()
+                await self.close(reason=f"failure-to-start-{str(type(exc))}")
                 self.status = Status.failed
                 self.__startup_exc = exc
 
             try:
-                await asyncio.wait_for(self.start_unsafe(), timeout=timeout)
+                await wait_for(self.start_unsafe(), timeout=timeout)
             except asyncio.TimeoutError as exc:
                 await _close_on_failure(exc)
                 raise asyncio.TimeoutError(
@@ -534,7 +670,8 @@ class Server:
             except Exception as exc:
                 await _close_on_failure(exc)
                 raise RuntimeError(f"{type(self).__name__} failed to start.") from exc
-            self.status = Status.running
+            if self.status == Status.init:
+                self.status = Status.running
         return self
 
     async def __aenter__(self):
@@ -559,26 +696,38 @@ class Server:
             if not pc.is_running():
                 pc.start()
 
-    def stop(self):
-        if self.__stopped:
-            return
+    def _stop_listeners(self) -> asyncio.Future:
+        listeners_to_stop: set[Awaitable] = set()
 
-        self.__stopped = True
-        _stops = set()
         for listener in self.listeners:
             future = listener.stop()
             if inspect.isawaitable(future):
-                _stops.add(future)
+                warnings.warn(
+                    f"{type(listener)} is using an asynchronous `stop` method. "
+                    "Support for asynchronous `Listener.stop` has been deprecated and "
+                    "will be removed in a future version",
+                    DeprecationWarning,
+                )
+                listeners_to_stop.add(future)
+            elif hasattr(listener, "abort_handshaking_comms"):
+                listener.abort_handshaking_comms()
 
-        if _stops:
+        return asyncio.gather(*listeners_to_stop)
 
-            async def background_stops():
-                await asyncio.gather(*_stops)
-
-            self._ongoing_background_tasks.call_soon(background_stops)
+    def stop(self) -> None:
+        if self.__stopped:
+            return
+        self.__stopped = True
+        self.monitor.close()
+        if not (stop_listeners := self._stop_listeners()).done():
+            self._ongoing_background_tasks.call_soon(
+                asyncio.wait_for(stop_listeners, timeout=None)  # type: ignore[arg-type]
+            )
+        if self._workdir is not None:
+            self._workdir.release()
 
     @property
-    def listener(self):
+    def listener(self) -> Listener | None:
         if self.listeners:
             return self.listeners[0]
         else:
@@ -621,6 +770,7 @@ class Server:
             if self.listener is None:
                 raise ValueError("cannot get address of non-running Server")
             self._address = self.listener.contact_address
+            assert self._address
         return self._address
 
     @property
@@ -673,7 +823,7 @@ class Server:
     def identity(self) -> dict[str, str]:
         return {"type": type(self).__name__, "id": self.id}
 
-    def _to_dict(self, *, exclude: Container[str] = ()) -> dict:
+    def _to_dict(self, *, exclude: Container[str] = ()) -> dict[str, Any]:
         """Dictionary representation for debugging purposes.
         Not type stable and not intended for roundtrips.
 
@@ -683,7 +833,7 @@ class Server:
         Client.dump_cluster_state
         distributed.utils.recursive_to_dict
         """
-        info = self.identity()
+        info: dict[str, Any] = self.identity()
         extra = {
             "address": self.address,
             "status": self.status.name,
@@ -715,7 +865,7 @@ class Server:
         )
         self.listeners.append(listener)
 
-    def handle_comm(self, comm):
+    def handle_comm(self, comm: Comm) -> NoOpAwaitable:
         """Start a background task that dispatches new communications to coroutine-handlers"""
         try:
             self._ongoing_background_tasks.call_soon(self._handle_comm, comm)
@@ -723,7 +873,7 @@ class Server:
             comm.abort()
         return NoOpAwaitable()
 
-    async def _handle_comm(self, comm):
+    async def _handle_comm(self, comm: Comm) -> None:
         """Dispatch new communications to coroutine-handlers
 
         Handlers is a dictionary mapping operation names to functions or
@@ -750,7 +900,7 @@ class Server:
                     msg = await comm.read()
                     logger.debug("Message from %r: %s", address, msg)
                 except OSError as e:
-                    if not sys.is_finalizing():
+                    if not is_python_shutting_down():
                         logger.debug(
                             "Lost connection to %r while reading message: %s."
                             " Last operation: %s",
@@ -854,7 +1004,7 @@ class Server:
 
         finally:
             del self._comms[comm]
-            if not sys.is_finalizing() and not comm.closed():
+            if not is_python_shutting_down() and not comm.closed():
                 try:
                     comm.abort()
                 except Exception as e:
@@ -862,7 +1012,9 @@ class Server:
                         "Failed while closing connection to %r: %s", address, e
                     )
 
-    async def handle_stream(self, comm, extra=None):
+    async def handle_stream(
+        self, comm: Comm, extra: dict[str, Any] | None = None
+    ) -> None:
         extra = extra or {}
         logger.info("Starting established connection to %s", comm.peer_address)
 
@@ -896,10 +1048,7 @@ class Server:
                             break
                         handler = self.stream_handlers[op]
                         if iscoroutinefunction(handler):
-                            self._ongoing_background_tasks.call_soon(
-                                handler, **merge(extra, msg)
-                            )
-                            await asyncio.sleep(0)
+                            await handler(**merge(extra, msg))
                         else:
                             handler(**merge(extra, msg))
                     else:
@@ -915,44 +1064,63 @@ class Server:
             await comm.close()
             assert comm.closed()
 
-    async def close(self, timeout=None):
+    async def close(self, timeout: float | None = None, reason: str = "") -> None:
         try:
             for pc in self.periodic_callbacks.values():
                 pc.stop()
 
-            if not self.__stopped:
-                self.__stopped = True
-                _stops = set()
-                for listener in self.listeners:
-                    future = listener.stop()
-                    if inspect.isawaitable(future):
-                        warnings.warn(
-                            f"{type(listener)} is using an asynchronous `stop` method. "
-                            "Support for asynchronous `Listener.stop` will be removed in a future version",
-                            PendingDeprecationWarning,
-                        )
-                        _stops.add(future)
-                if _stops:
-                    await asyncio.gather(*_stops)
+            self.__stopped = True
+            self.monitor.close()
+            await self._stop_listeners()
 
             # TODO: Deal with exceptions
             await self._ongoing_background_tasks.stop()
 
             await self.rpc.close()
             await asyncio.gather(*[comm.close() for comm in list(self._comms)])
+
+            # Remove scratch directory from global sys.path
+            if self._updated_sys_path and sys.path[0] == self.local_directory:
+                sys.path.remove(self.local_directory)
         finally:
             self._event_finished.set()
 
-    def digest_metric(self, name: str, value: float) -> None:
+    def digest_metric(self, name: Hashable, value: float) -> None:
         # Granular data (requires crick)
         if self.digests is not None:
             self.digests[name].add(value)
         # Cumulative data (reset by server restart)
         self.digests_total[name] += value
+        # Cumulative data sent to scheduler and reset on heartbeat
+        self.digests_total_since_heartbeat[name] += value
         # Local maximums (reset by Prometheus poll)
         self.digests_max[name] = max(self.digests_max[name], value)
 
 
+def context_meter_to_server_digest(digest_tag: str) -> Callable:
+    """Decorator for an async method of a Server subclass that calls
+    ``distributed.metrics.context_meter.meter`` and/or ``digest_metric``.
+    It routes the calls from ``context_meter.digest_metric(label, value, unit)`` to
+    ``Server.digest_metric((digest_tag, label, unit), value)``.
+    """
+
+    def decorator(func: Callable) -> Callable:
+        @wraps(func)
+        async def wrapper(self: Server, *args: Any, **kwargs: Any) -> Any:
+            def metrics_callback(label: Hashable, value: float, unit: str) -> None:
+                if not isinstance(label, tuple):
+                    label = (label,)
+                name = (digest_tag, *label, unit)
+                self.digest_metric(name, value)
+
+            with context_meter.add_callback(metrics_callback, allow_offload=True):
+                return await func(self, *args, **kwargs)
+
+        return wrapper
+
+    return decorator
+
+
 def pingpong(comm):
     return b"pong"
 
@@ -1014,13 +1182,19 @@ async def send_recv(  # type: ignore[no-untyped-def]
     return response
 
 
-def addr_from_args(addr=None, ip=None, port=None):
+def addr_from_args(
+    addr: str | tuple[str, int | None] | None = None,
+    ip: str | None = None,
+    port: int | None = None,
+) -> str:
     if addr is None:
-        addr = (ip, port)
-    else:
-        assert ip is None and port is None
+        assert ip is not None
+        return normalize_address(unparse_host_port(ip, port))
+
+    assert ip is None and port is None
     if isinstance(addr, tuple):
-        addr = unparse_host_port(*addr)
+        return normalize_address(unparse_host_port(*addr))
+
     return normalize_address(addr)
 
 
@@ -1288,20 +1462,20 @@ class ConnectionPool:
 
     def __init__(
         self,
-        limit=512,
-        deserialize=True,
-        serializers=None,
-        allow_offload=True,
-        deserializers=None,
-        connection_args=None,
-        timeout=None,
-        server=None,
-    ):
+        limit: int = 512,
+        deserialize: bool = True,
+        serializers: list[str] | None = None,
+        allow_offload: bool = True,
+        deserializers: list[str] | None = None,
+        connection_args: dict[str, object] | None = None,
+        timeout: float | None = None,
+        server: object = None,
+    ) -> None:
         self.limit = limit  # Max number of open comms
         # Invariant: len(available) == open - active
-        self.available = defaultdict(set)
+        self.available: defaultdict[str, set[Comm]] = defaultdict(set)
         # Invariant: len(occupied) == active
-        self.occupied = defaultdict(set)
+        self.occupied: defaultdict[str, set[Comm]] = defaultdict(set)
         self.allow_offload = allow_offload
         self.deserialize = deserialize
         self.serializers = serializers
@@ -1309,18 +1483,20 @@ class ConnectionPool:
         self.connection_args = connection_args or {}
         self.timeout = timeout
         self.server = weakref.ref(server) if server else None
-        self._created = weakref.WeakSet()
+        self._created: weakref.WeakSet[Comm] = weakref.WeakSet()
         self._instances.add(self)
         # _n_connecting and _connecting have subtle different semantics. The set
         # _connecting contains futures actively trying to establish a connection
         # while the _n_connecting also accounts for connection attempts which
         # are waiting due to the connection limit
-        self._connecting = set()
+        self._connecting: defaultdict[str, set[Callable[[str], None]]] = defaultdict(
+            set
+        )
         self._pending_count = 0
         self._connecting_count = 0
         self.status = Status.init
 
-    def _validate(self):
+    def _validate(self) -> None:
         """
         Validate important invariants of this class
 
@@ -1329,35 +1505,48 @@ class ConnectionPool:
         assert self.semaphore._value == self.limit - self.open - self._n_connecting
 
     @property
-    def active(self):
+    def active(self) -> int:
         return sum(map(len, self.occupied.values()))
 
     @property
-    def open(self):
+    def open(self) -> int:
         return self.active + sum(map(len, self.available.values()))
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "<ConnectionPool: open=%d, active=%d, connecting=%d>" % (
             self.open,
             self.active,
             len(self._connecting),
         )
 
-    def __call__(self, addr=None, ip=None, port=None):
+    def __call__(
+        self,
+        addr: str | tuple[str, int | None] | None = None,
+        ip: str | None = None,
+        port: int | None = None,
+    ) -> PooledRPCCall:
         """Cached rpc objects"""
         addr = addr_from_args(addr=addr, ip=ip, port=port)
         return PooledRPCCall(
             addr, self, serializers=self.serializers, deserializers=self.deserializers
         )
 
-    def __await__(self):
-        async def _():
+    def __await__(self) -> Generator[Any, Any, Self]:
+        async def _() -> Self:
             await self.start()
             return self
 
         return _().__await__()
 
-    async def start(self):
+    async def __aenter__(self):
+        await self
+        return self
+
+    async def __aexit__(self, *args):
+        await self.close()
+        return
+
+    async def start(self) -> None:
         # Invariant: semaphore._value == limit - open - _n_connecting
         self.semaphore = asyncio.Semaphore(self.limit)
         self.status = Status.running
@@ -1366,7 +1555,7 @@ class ConnectionPool:
     def _n_connecting(self) -> int:
         return self._connecting_count
 
-    async def _connect(self, addr, timeout=None):
+    async def _connect(self, addr: str, timeout: float | None = None) -> Comm:
         self._pending_count += 1
         try:
             await self.semaphore.acquire()
@@ -1391,15 +1580,15 @@ class ConnectionPool:
                 raise
             finally:
                 self._connecting_count -= 1
-        except asyncio.CancelledError:
-            raise CommClosedError("ConnectionPool closing.")
         finally:
             self._pending_count -= 1
 
-    async def connect(self, addr, timeout=None):
+    async def connect(self, addr: str, timeout: float | None = None) -> Comm:
         """
         Get a Comm to the given address.  For internal use.
         """
+        if self.status != Status.running:
+            raise RuntimeError("ConnectionPool is closed")
         available = self.available[addr]
         occupied = self.occupied[addr]
         while available:
@@ -1413,32 +1602,80 @@ class ConnectionPool:
         if self.semaphore.locked():
             self.collect()
 
-        # This construction is there to ensure that cancellation requests from
-        # the outside can be distinguished from cancellations of our own.
-        # Once the CommPool closes, we'll cancel the connect_attempt which will
-        # raise an OSError
-        # If the ``connect`` is cancelled from the outside, the Event.wait will
-        # be cancelled instead which we'll reraise as a CancelledError and allow
-        # it to propagate
-        connect_attempt = asyncio.create_task(self._connect(addr, timeout))
-        done = asyncio.Event()
-        self._connecting.add(connect_attempt)
-        connect_attempt.add_done_callback(lambda _: done.set())
-        connect_attempt.add_done_callback(self._connecting.discard)
-
-        try:
-            await done.wait()
-        except asyncio.CancelledError:
-            # This is an outside cancel attempt
-            connect_attempt.cancel()
+        # on 3.11 this uses asyncio.timeout as a cancel scope to avoid having
+        # to track inner and outer CancelledError exceptions and correctly
+        # call .uncancel() when a CancelledError is caught.
+        if sys.version_info >= (3, 11):
+            reason: str | None = None
             try:
-                await connect_attempt
-            except CommClosedError:
-                pass
-            raise
-        return await connect_attempt
+                async with asyncio.timeout(math.inf) as scope:
 
-    def reuse(self, addr, comm):
+                    def cancel_timeout_cb(reason_: str) -> None:
+                        nonlocal reason
+                        reason = reason_
+                        scope.reschedule(-1)
+
+                    self._connecting[addr].add(cancel_timeout_cb)
+                    try:
+                        return await self._connect(addr=addr, timeout=timeout)
+                    finally:
+                        connecting = self._connecting[addr]
+                        connecting.discard(cancel_timeout_cb)
+                        if not connecting:
+                            try:
+                                del self._connecting[addr]
+                            except KeyError:
+                                pass
+            except TimeoutError:
+                if reason is None:
+                    raise
+                raise CommClosedError(reason)
+        else:
+            # This construction is there to ensure that cancellation requests from
+            # the outside can be distinguished from cancellations of our own.
+            # Once the CommPool closes, we'll cancel the connect_attempt which will
+            # raise an OSError
+            # If the ``connect`` is cancelled from the outside, the Event.wait will
+            # be cancelled instead which we'll reraise as a CancelledError and allow
+            # it to propagate
+            connect_attempt = asyncio.create_task(self._connect(addr, timeout))
+            done = asyncio.Event()
+            reason = "ConnectionPool closing."
+
+            def cancel_task_cb(reason_: str) -> None:
+                nonlocal reason
+                reason = reason_
+                connect_attempt.cancel()
+
+            connecting = self._connecting[addr]
+            connecting.add(cancel_task_cb)
+
+            def callback(task: asyncio.Task[Comm]) -> None:
+                done.set()
+                connecting = self._connecting[addr]
+                connecting.discard(cancel_task_cb)
+
+                if not connecting:
+                    try:
+                        del self._connecting[addr]
+                    except KeyError:  # pragma: no cover
+                        pass
+
+            connect_attempt.add_done_callback(callback)
+
+            try:
+                await done.wait()
+            except asyncio.CancelledError:
+                # This is an outside cancel attempt
+                connect_attempt.cancel()
+                await connect_attempt
+                raise
+            try:
+                return connect_attempt.result()
+            except asyncio.CancelledError:
+                raise CommClosedError(reason)
+
+    def reuse(self, addr: str, comm: Comm) -> None:
         """
         Reuse an open communication to the given address.  For internal use.
         """
@@ -1457,7 +1694,7 @@ class ConnectionPool:
                 if self.semaphore.locked() and self._pending_count:
                     self.collect()
 
-    def collect(self):
+    def collect(self) -> None:
         """
         Collect open but unused communications, to allow opening other ones.
         """
@@ -1473,11 +1710,11 @@ class ConnectionPool:
                 self.semaphore.release()
             comms.clear()
 
-    def remove(self, addr):
+    def remove(self, addr: str, *, reason: str = "Address removed.") -> None:
         """
         Remove all Comms to a given address.
         """
-        logger.info("Removing comms to %s", addr)
+        logger.debug("Removing comms to %s", addr)
         if addr in self.available:
             comms = self.available.pop(addr)
             for comm in comms:
@@ -1489,13 +1726,19 @@ class ConnectionPool:
                 IOLoop.current().add_callback(comm.close)
                 self.semaphore.release()
 
-    async def close(self):
+        if addr in self._connecting:
+            cbs = self._connecting[addr]
+            for cb in cbs:
+                cb(reason)
+
+    async def close(self) -> None:
         """
         Close all communications
         """
         self.status = Status.closed
-        for conn_fut in self._connecting:
-            conn_fut.cancel()
+        for cbs in self._connecting.values():
+            for cb in cbs:
+                cb("ConnectionPool closing.")
         for d in [self.available, self.occupied]:
             comms = set()
             while d:
@@ -1535,6 +1778,10 @@ class ErrorMessage(TypedDict):
     traceback_text: str
 
 
+class OKMessage(TypedDict):
+    status: Literal["OK"]
+
+
 def error_message(e: BaseException, status: str = "error") -> ErrorMessage:
     """Produce message to send back given an exception has occurred
 
