@@ -8,14 +8,16 @@ from datetime import timedelta
 from typing import TYPE_CHECKING, cast
 
 import tlz as toolz
-from tornado.ioloop import IOLoop, PeriodicCallback
+from tornado.ioloop import IOLoop
 
+import dask.config
 from dask.utils import parse_timedelta
 
-from ..metrics import time
+from distributed.compatibility import PeriodicCallback
+from distributed.metrics import time
 
 if TYPE_CHECKING:
-    from ..scheduler import WorkerState
+    from distributed.scheduler import WorkerState
 
 
 logger = logging.getLogger(__name__)
@@ -94,7 +96,7 @@ class AdaptiveCore:
         minimum: int = 0,
         maximum: int | float = math.inf,
         wait_count: int = 3,
-        interval: str | int | float | timedelta | None = "1s",
+        interval: str | int | float | timedelta = "1s",
     ):
         if not isinstance(maximum, int) and not math.isinf(maximum):
             raise TypeError(f"maximum must be int or inf; got {maximum}")
@@ -134,7 +136,9 @@ class AdaptiveCore:
         # internal state
         self.close_counts = defaultdict(int)
         self._adapting = False
-        self.log = deque(maxlen=10000)
+        self.log = deque(
+            maxlen=dask.config.get("distributed.admin.low-level-log-length")
+        )
 
     def stop(self) -> None:
         logger.info("Adaptive stop")
@@ -224,7 +228,6 @@ class AdaptiveCore:
         status = None
 
         try:
-
             target = await self.safe_target()
             recommendations = await self.recommendations(target)
 
