@@ -9,14 +9,13 @@ from dask import delayed
 
 import distributed
 from distributed import Client, Event, Future, Worker, span, wait
-from distributed.compatibility import WINDOWS
 from distributed.diagnostics.plugin import SchedulerPlugin
-from distributed.metrics import time
 from distributed.utils_test import (
     NoSchedulerDelayWorker,
     async_poll_for,
     gen_cluster,
     inc,
+    padded_time,
     slowinc,
     wait_for_state,
 )
@@ -26,25 +25,22 @@ from distributed.utils_test import (
 async def test_spans(c, s, a):
     x = delayed(inc)(1)  # Default span
 
-    @span("p2")
-    def f(i):
-        return i * 2
-
     with span("my workflow") as mywf_id:
         with span("p1") as p1_id:
             y = x + 1
-        z = f(y)
+        with span("p2") as p2_id:
+            z = y * 2
 
     zp = c.persist(z)
     assert await c.compute(zp) == 6
 
     ext = s.extensions["spans"]
 
-    p2_id = s.tasks[z.key].group.span_id
     assert mywf_id
     assert p1_id
     assert p2_id
     assert s.tasks[y.key].group.span_id == p1_id
+    assert s.tasks[z.key].group.span_id == p2_id
     assert mywf_id != p1_id != p2_id
 
     expect_annotations = {
@@ -135,15 +131,12 @@ async def test_repeat_span(c, s, a, b):
     """Opening and closing the same span will result in multiple spans with different
     ids and same name
     """
-
-    @span("foo")
-    def f(x, key):
-        return c.submit(inc, x, key=key)
-
     with span("foo"):
         x = c.submit(inc, 1, key="x")
-    y = f(x, key="y")
-    z = f(y, key="z")
+    with span("foo"):
+        y = c.submit(inc, x, key="y")
+    with span("foo"):
+        z = c.submit(inc, y, key="z")
     assert await z == 4
 
     sbn = s.extensions["spans"].spans_search_by_name["foo",]
@@ -220,8 +213,10 @@ async def test_no_extension(c, s, a, b):
     Worker=NoSchedulerDelayWorker,
     config={"optimization.fuse.active": False},
 )
-async def test_task_groups(c, s, a, b, release):
+async def test_task_groups(c, s, a, b, release, no_time_resync):
     da = pytest.importorskip("dask.array")
+    t0 = await padded_time(before=0)
+
     with span("wf"):
         with span("p1"):
             a = da.ones(10, chunks=5, dtype="int64") + 1
@@ -230,9 +225,8 @@ async def test_task_groups(c, s, a, b, release):
         a = a.sum()  # A TaskGroup attached directly to a non-leaf Span
         finalizer = c.compute(a)
 
-    t0 = time()
     assert await finalizer == 40
-    t1 = time()
+    t1 = await padded_time(after=0)
 
     if release:
         # Test that the information in the Spans survives the tasks
@@ -289,13 +283,13 @@ async def test_task_groups(c, s, a, b, release):
 
 
 @gen_cluster(client=True, nthreads=[("", 1)], Worker=NoSchedulerDelayWorker)
-async def test_before_first_task_finished(c, s, a):
-    t0 = time()
+async def test_before_first_task_finished(c, s, a, no_time_resync):
+    t0 = await padded_time(before=0)
     ev = Event()
     x = c.submit(ev.wait, key="x")
     await wait_for_state("x", "executing", a)
     sp = s.extensions["spans"].spans_search_by_name["default",][-1]
-    t1 = time()
+    t1 = await padded_time()
     assert t0 < sp.enqueued < t1
     assert sp.start == 0
     assert t1 < sp.stop < t1 + 1
@@ -305,7 +299,7 @@ async def test_before_first_task_finished(c, s, a):
 
     await ev.set()
     await x
-    t2 = time()
+    t2 = await padded_time()
     assert t0 < sp.enqueued < sp.start < t1 < sp.stop < t2
     assert sp.duration > 0
     assert sp.all_durations["compute"] > 0
@@ -363,10 +357,10 @@ async def test_mismatched_span(c, s, a, use_default):
     assert len(sbn[span_name][0].groups) == 1
     assert s.task_groups["x"].span_id == sbn[span_name][0].id
 
-    sts0 = s.tasks[str(x0.key)]
-    sts1 = s.tasks[str(x1.key)]
-    wts0 = a.state.tasks[str(x0.key)]
-    wts1 = a.state.tasks[str(x1.key)]
+    sts0 = s.tasks[x0.key]
+    sts1 = s.tasks[x1.key]
+    wts0 = a.state.tasks[x0.key]
+    wts1 = a.state.tasks[x1.key]
     assert sts0.group is sts1.group
     assert wts0.span_id == wts1.span_id
 
@@ -377,8 +371,8 @@ async def test_mismatched_span(c, s, a, use_default):
     else:
         expect = {"ids": (wts0.span_id,), "name": ("p1",)}
         assert s.plugins["my-plugin"].annotations == [
-            {"span": {"('x', 0)": expect}},
-            {"span": {"('x', 1)": expect}},
+            {"span": {("x", 0): expect}},
+            {"span": {("x", 1): expect}},
         ]
         for ts in (sts0, sts1, wts0, wts1):
             assert ts.annotations["span"] == expect
@@ -518,14 +512,12 @@ async def test_worker_metrics(c, s, a, b):
 
     # metrics for foo include self and its child bar
     assert list(foo_metrics) == [
-        ("execute", "x", "deserialize", "seconds"),
         ("execute", "x", "thread-cpu", "seconds"),
         ("execute", "x", "thread-noncpu", "seconds"),
         ("execute", "x", "executor", "seconds"),
         ("execute", "x", "other", "seconds"),
         ("execute", "x", "memory-read", "count"),
         ("execute", "x", "memory-read", "bytes"),
-        ("execute", "y", "deserialize", "seconds"),
         ("execute", "y", "thread-cpu", "seconds"),
         ("execute", "y", "thread-noncpu", "seconds"),
         ("execute", "y", "executor", "seconds"),
@@ -536,7 +528,6 @@ async def test_worker_metrics(c, s, a, b):
         list(bar0_metrics)
         == list(bar1_metrics)
         == [
-            ("execute", "y", "deserialize", "seconds"),
             ("execute", "y", "thread-cpu", "seconds"),
             ("execute", "y", "thread-noncpu", "seconds"),
             ("execute", "y", "executor", "seconds"),
@@ -545,9 +536,13 @@ async def test_worker_metrics(c, s, a, b):
         ]
     )
 
-    if not WINDOWS:
-        for metrics in (foo_metrics, bar0_metrics, bar1_metrics):
-            assert all(v > 0 for v in metrics.values()), metrics
+    for metrics in (foo_metrics, bar0_metrics, bar1_metrics):
+        # On OSes other than Windows, time metrics are, 99.9% of the time, strictly
+        # greater than zero. However adjustments in the wall clock can occasionally
+        # cause them to be exactly zero. Here we're testing that there is a floor to
+        # zero implemented.
+        for k, v in metrics.items():
+            assert v >= 0, (k, v)
 
     # Metrics have been synchronized from scheduler to spans
     for k, v in foo_metrics.items():
@@ -644,7 +639,6 @@ async def test_code(c, s, a, b):
 
     # Identical code stacks have been deduplicated
     assert len(code) == 3
-    assert "def _run(self)" in code[0][-2].code
     assert "inc, 1" in code[0][-1].code
     assert code[0][-1].lineno_relative == 10
     assert code[1][-1].lineno_relative == 11
@@ -680,7 +674,7 @@ async def test_merge_nothing(s):
 
 @gen_cluster(client=True)
 async def test_active_cpu_seconds_trivial(c, s, a, b):
-    await c.submit(slowinc, 1, delay=0.1, key="x")
+    await c.submit(slowinc, 1, delay=0.2, key="x")
     await a.heartbeat()
     await b.heartbeat()
     span = s.extensions["spans"].spans_search_by_name["default",][0]
@@ -688,13 +682,17 @@ async def test_active_cpu_seconds_trivial(c, s, a, b):
     assert span.done
     assert span.nthreads_intervals == [(span.enqueued, span.stop, 3)]
     assert span.active_cpu_seconds == (span.stop - span.enqueued) * 3
-    k = "execute", "N/A", "idle or other spans", "seconds"
-    assert 0.15 < span.cumulative_worker_metrics[k] < span.active_cpu_seconds
+    idle_time = span.cumulative_worker_metrics[
+        "execute", "N/A", "idle or other spans", "seconds"
+    ]
+    # Allow for some very generous margins to compensate for occasional corrections in
+    # the wall clock (much more frequent in Windows)
+    assert 0.15 < idle_time <= span.active_cpu_seconds
 
 
 @pytest.mark.parametrize("some_done", [False, True])
 @gen_cluster(client=True, nthreads=[("", 2)], Worker=NoSchedulerDelayWorker)
-async def test_active_cpu_seconds_not_done(c, s, a, some_done):
+async def test_active_cpu_seconds_not_done(c, s, a, some_done, no_time_resync):
     ev = Event()
     x0 = c.submit(ev.wait, key="x-0", workers=[a.address])
     if some_done:
@@ -704,7 +702,7 @@ async def test_active_cpu_seconds_not_done(c, s, a, some_done):
 
     span = s.extensions["spans"].spans_search_by_name["default",][0]
     assert not span.done
-    now = time()
+    now = await padded_time()
 
     intervals = span.nthreads_intervals
     assert len(intervals) == 1
@@ -722,7 +720,7 @@ async def test_active_cpu_seconds_not_done(c, s, a, some_done):
 
 
 @gen_cluster(client=True, Worker=NoSchedulerDelayWorker)
-async def test_active_cpu_seconds_change_nthreads(c, s, a, b):
+async def test_active_cpu_seconds_change_nthreads(c, s, a, b, no_time_resync):
     ev = Event()
     x = c.submit(ev.wait, key="x", workers=[a.address])
     await wait_for_state("x", "executing", a)
@@ -775,7 +773,7 @@ async def test_active_cpu_seconds_change_nthreads(c, s, a, b):
 
 
 @gen_cluster(client=True, nthreads=[("", 2)], Worker=NoSchedulerDelayWorker)
-async def test_active_cpu_seconds_merged(c, s, a):
+async def test_active_cpu_seconds_merged(c, s, a, no_time_resync):
     """Overlapping input spans are not double-counted
     Empty gap between input spans is not counted
     """
@@ -843,3 +841,44 @@ async def test_spans_are_visible_from_tasks(c, s, a, b):
 
     # No annotation is created for the default span
     assert await c.submit(dask.get_annotations) == {}
+
+
+@gen_cluster(client=True)
+async def test_span_on_persist(c, s, a, b):
+    """As a workaround to lack of annotations support in dask-expr and loss of
+    annotations due to low level optimization in dask.array, you can use span() to wrap
+    calls to persist() and compute()
+    """
+    x = delayed(inc)(1, dask_key_name="x")
+    with span("x") as x_id:
+        x = c.persist(x)
+    y = delayed(inc)(x, dask_key_name="y")
+    with span("y") as y_id:
+        y = c.compute(y)
+    assert await y == 3
+
+    assert s.tasks["x"].group.span_id == x_id
+    assert s.tasks["y"].group.span_id == y_id
+
+
+@pytest.mark.filterwarnings("ignore:Dask annotations")
+@gen_cluster(client=True)
+async def test_collections_metadata(c, s, a, b):
+    pd = pytest.importorskip("pandas")
+    dd = pytest.importorskip("dask.dataframe")
+    np = pytest.importorskip("numpy")
+    df = pd.DataFrame(
+        {"x": np.random.random(1000), "y": np.random.random(1000)},
+        index=np.arange(1000),
+    )
+    ldf = dd.from_pandas(df, npartitions=10)
+
+    with span("foo") as span_id:
+        await c.compute(ldf)
+
+    ext = s.extensions["spans"]
+    span_ = ext.spans[span_id]
+    collections_meta = span_.metadata["collections"]
+    assert isinstance(collections_meta, list)
+    assert len(collections_meta) == 1
+    assert collections_meta[0]["type"] == type(ldf).__name__
