@@ -9,12 +9,14 @@ import multiprocessing
 import os
 import queue
 import socket
+import sys
 import traceback
 import warnings
 import xml
 from array import array
 from collections import deque
 from concurrent.futures import Executor, Future, ThreadPoolExecutor
+from contextvars import ContextVar
 from time import sleep
 from unittest import mock
 
@@ -96,16 +98,21 @@ async def test_All():
         assert end - start < 10
 
 
+def test_sync(loop_in_thread):
+    async def f(x, y):
+        await asyncio.sleep(0.01)
+        return x, y
+
+    result = sync(loop_in_thread, f, 1, y=2)
+    assert result == (1, 2)
+
+
 def test_sync_error(loop_in_thread):
-    loop = loop_in_thread
-    try:
-        result = sync(loop, throws, 1)
-    except Exception as exc:
-        f = exc
-        assert "hello" in str(exc)
-        tb = get_traceback()
-        L = traceback.format_tb(tb)
-        assert any("throws" in line for line in L)
+    with pytest.raises(RuntimeError, match="hello!") as exc:
+        sync(loop_in_thread, throws, 1)
+
+    L = traceback.format_tb(exc.value.__traceback__)
+    assert any("throws" in line for line in L)
 
     def function1(x):
         return function2(x)
@@ -113,18 +120,15 @@ def test_sync_error(loop_in_thread):
     def function2(x):
         return throws(x)
 
-    try:
-        result = sync(loop, function1, 1)
-    except Exception as exc:
-        assert "hello" in str(exc)
-        tb = get_traceback()
-        L = traceback.format_tb(tb)
-        assert any("function1" in line for line in L)
-        assert any("function2" in line for line in L)
+    with pytest.raises(RuntimeError, match="hello!") as exc:
+        sync(loop_in_thread, function1, 1)
+
+    L = traceback.format_tb(exc.value.__traceback__)
+    assert any("function1" in line for line in L)
+    assert any("function2" in line for line in L)
 
 
 def test_sync_timeout(loop_in_thread):
-    loop = loop_in_thread
     with pytest.raises(TimeoutError):
         sync(loop_in_thread, asyncio.sleep, 0.5, callback_timeout=0.05)
 
@@ -144,6 +148,21 @@ def test_sync_closed_loop():
     exc_info.match("IOLoop is clos(ed|ing)")
 
 
+def test_sync_contextvars(loop_in_thread):
+    """Test that sync() propagates contextvars - namely,
+    distributed.metrics.context_meter callbacks
+    """
+    v = ContextVar("v", default=0)
+
+    async def f():
+        return v.get()
+
+    assert sync(loop_in_thread, f) == 0
+    tok = v.set(1)
+    assert sync(loop_in_thread, f) == 1
+    v.reset(tok)
+
+
 def test_is_kernel():
     pytest.importorskip("IPython")
     assert is_kernel() is False
@@ -700,58 +719,10 @@ async def test_offload_preserves_contextvars():
     await asyncio.gather(set_var("foo"), set_var("bar"))
 
 
-def test_serialize_for_cli_deprecated():
-    with pytest.warns(FutureWarning, match="serialize_for_cli is deprecated"):
-        from distributed.utils import serialize_for_cli
-    assert serialize_for_cli is dask.config.serialize
-
-
-def test_deserialize_for_cli_deprecated():
-    with pytest.warns(FutureWarning, match="deserialize_for_cli is deprecated"):
-        from distributed.utils import deserialize_for_cli
-    assert deserialize_for_cli is dask.config.deserialize
-
-
-def test_parse_bytes_deprecated():
-    with pytest.warns(FutureWarning, match="parse_bytes is deprecated"):
-        from distributed.utils import parse_bytes
-    assert parse_bytes is dask.utils.parse_bytes
-
-
-def test_format_bytes_deprecated():
-    with pytest.warns(FutureWarning, match="format_bytes is deprecated"):
-        from distributed.utils import format_bytes
-    assert format_bytes is dask.utils.format_bytes
-
-
-def test_format_time_deprecated():
-    with pytest.warns(FutureWarning, match="format_time is deprecated"):
-        from distributed.utils import format_time
-    assert format_time is dask.utils.format_time
-
-
-def test_funcname_deprecated():
-    with pytest.warns(FutureWarning, match="funcname is deprecated"):
-        from distributed.utils import funcname
-    assert funcname is dask.utils.funcname
-
-
-def test_parse_timedelta_deprecated():
-    with pytest.warns(FutureWarning, match="parse_timedelta is deprecated"):
-        from distributed.utils import parse_timedelta
-    assert parse_timedelta is dask.utils.parse_timedelta
-
-
-def test_typename_deprecated():
-    with pytest.warns(FutureWarning, match="typename is deprecated"):
-        from distributed.utils import typename
-    assert typename is dask.utils.typename
-
-
-def test_tmpfile_deprecated():
-    with pytest.warns(FutureWarning, match="tmpfile is deprecated"):
-        from distributed.utils import tmpfile
-    assert tmpfile is dask.utils.tmpfile
+def test_no_default_deprecated():
+    with pytest.warns(FutureWarning, match="no_default is deprecated"):
+        from distributed.utils import no_default
+    assert no_default is dask.typing.no_default
 
 
 def test_iscoroutinefunction_unhashable_input():
@@ -908,7 +879,10 @@ async def test_log_errors():
         pass
 
     # Use the logger of the caller module
-    with captured_logger("test_utils") as caplog:
+    with (
+        captured_logger("test_utils") as caplog,
+        captured_logger("distributed.utils") as caplog2,
+    ):
         # Context manager
         with log_errors():
             pass
@@ -935,6 +909,8 @@ async def test_log_errors():
             return 123
 
         assert _() == 123
+        with pytest.raises(TypeError):
+            _(bad=1)
 
         @log_errors
         def _():
@@ -982,6 +958,9 @@ async def test_log_errors():
 
         assert await _() == 123
 
+        task = asyncio.create_task(_())
+        assert await task == 123
+
         @log_errors
         async def _():
             raise CustomError("err6")
@@ -989,6 +968,29 @@ async def test_log_errors():
         with pytest.raises(CustomError):
             await _()
 
+        task = asyncio.create_task(_())
+        with pytest.raises(CustomError):
+            await task
+
+        @log_errors()
+        async def _():
+            raise CustomError("err7")
+
+        with pytest.raises(CustomError):
+            await _()
+
+        task = asyncio.create_task(_())
+        with pytest.raises(CustomError):
+            await task
+
+        async def _():
+            with log_errors():
+                raise CustomError("err8")
+
+        task = asyncio.create_task(_())
+        with pytest.raises(CustomError):
+            await task
+
     assert [row for row in caplog.getvalue().splitlines() if row.startswith("err")] == [
         "err1",
         "err2",
@@ -999,15 +1001,36 @@ async def test_log_errors():
         "err5",
         "err5",
         "err6",
+        "err6",
+        "err7",
+        "err7",
+        "err8",
     ]
+    assert "got an unexpected keyword argument 'bad'" in caplog2.getvalue()
 
-    # Test unroll_stack
-    with captured_logger("distributed.utils") as caplog:
-        with pytest.raises(CustomError):
-            with log_errors(unroll_stack=0):
-                raise CustomError("err7")
 
-    assert caplog.getvalue().startswith("err7\n")
+@pytest.mark.parametrize("unroll_stack,logger_name", [(0, "test_utils"), (1, "a.b.c")])
+def test_log_errors_unroll_stack(unroll_stack, logger_name, tmp_path):
+    (tmp_path / "a").mkdir()
+    (tmp_path / "a").touch("__init__.py")
+    (tmp_path / "a" / "b").mkdir()
+    (tmp_path / "a" / "b").touch("__init__.py")
+    with (tmp_path / "a" / "b" / "c.py").open("w") as fh:
+        fh.write("def f():\n    raise ValueError('myerr')\n")
+
+    sys.path.insert(0, str(tmp_path))
+    import a.b.c
+
+    del sys.path[0]
+
+    with (
+        captured_logger(logger_name) as caplog,
+        pytest.raises(ValueError),
+        log_errors(unroll_stack=unroll_stack),
+    ):
+        a.b.c.f()
+
+    assert "myerr" in caplog.getvalue()
 
 
 def test_load_json_robust_timeout(tmp_path):
